// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_crud_bloc/src/core/enums/where_query_type.dart';
import 'package:wyatt_crud_bloc/src/core/model_identifier.dart';
import 'package:wyatt_crud_bloc/src/domain/entities/query.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/create.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/delete.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/delete_all.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/get.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/get_all.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/params/update_parameters.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/search.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/update.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/update_all.dart';
import 'package:wyatt_crud_bloc/src/features/crud/blocs/crud_base_cubit/crud_base_cubit.dart';

/// {@template crud_cubit_advanced}
/// Cubit that handles CRUD operations with more granularity.
/// {@endtemplate}
abstract class CrudAdvancedCubit<Model> extends CrudBaseCubit {
  /// {@macro crud_cubit}
  CrudAdvancedCubit() : super();

  Create<Model>? get crudCreate;
  DeleteAll<Model>? get crudDeleteAll;
  Delete<Model>? get crudDelete;
  GetAll<Model>? get crudGetAll;
  Get<Model>? get crudGet;
  Search<Model>? get crudSearch;
  UpdateAll<Model>? get crudUpdateAll;
  Update<Model>? get crudUpdate;

  /// Model identifier.
  /// Used to identify a model.
  ModelIdentifier<Model> get modelIdentifier;

  FutureOr<void> create(Model model) async {
    final crud = crudCreate;
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudCreating());
    final result = await crud.call(model);
    emit(
      result.fold(
        (_) {
          if (stateCopy is CrudLoaded<Model?>) {
            if (stateCopy.data == null) {
              return CrudLoaded<Model?>(model);
            }
            if (modelIdentifier.areEqual(stateCopy.data, model)) {
              return CrudLoaded<Model?>(model);
            }

            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            if (stateCopy.data.isEmpty) {
              return CrudListLoaded<Model?>([model]);
            }
            final List<Model?> lst = stateCopy.data.toList()..add(model);

            return CrudListLoaded<Model?>(lst);
          }

          return const CrudCreated();
        },
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> delete(String id) async {
    final crud = crudDelete;
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudDeleting());
    final result = await crud.call(id);
    emit(
      result.fold(
        (_) {
          if (stateCopy is CrudLoaded<Model?>) {
            if (modelIdentifier.isIdentifier(
              model: stateCopy.data,
              identifier: id,
            )) {
              return CrudLoaded<Model?>(null);
            }

            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            return CrudListLoaded<Model?>(
              stateCopy.data.where((element) {
                if (element == null) {
                  return false;
                }
                return !modelIdentifier.isIdentifier(
                  model: element,
                  identifier: id,
                );
              }).toList(),
            );
          }

          return const CrudDeleted();
        },
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> deleteAll() async {
    final crud = crudDeleteAll;
    if (crud == null) {
      return;
    }
    final stateCopy = state;
    emit(const CrudDeleting());
    final result = await crud.call(null);
    emit(
      result.fold(
        (_) {
          if (stateCopy is CrudLoaded<Model?>) {
            return CrudLoaded<Model?>(null);
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            return CrudListLoaded<Model?>(const []);
          }

          return const CrudOkReturn();
        },
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> get(String id) async {
    final crud = crudGet;
    if (crud == null) {
      return;
    }
    emit(const CrudReading());
    final result = await crud.call(id);
    emit(
      result.fold(
        CrudLoaded<Model?>.new,
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> getAll() async {
    final crud = crudGetAll;
    if (crud == null) {
      return;
    }
    emit(const CrudReading());
    final result = await crud.call(null);
    emit(
      result.fold(
        CrudListLoaded<Model?>.new,
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> search(List<Query> conditions) async {
    final crud = crudSearch;
    if (crud == null) {
      return;
    }

    emit(const CrudReading());
    final result = await crud.call(conditions);
    emit(
      result.fold(
        CrudListLoaded<Model?>.new,
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> update(UpdateParameters<Model> param) async {
    final crud = crudUpdate;
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudUpdating());
    final result = await crud.call(param);
    emit(
      await result.foldAsync(
        (_) async {
          if (stateCopy is CrudLoaded<Model?>) {
            if (modelIdentifier.isIdentifier(
              model: stateCopy.data,
              identifier: param.id,
            )) {
              // Same object, need to update actual stateCopy
              final crudGet = this.crudGet;
              if (crudGet == null) {
                // No read operation, can't update stateCopy.
                return stateCopy;
              }
              final newVersion = await crudGet.call(param.id);
              if (newVersion.isOk) {
                return CrudLoaded<Model?>(newVersion.ok);
              }
            }
            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            final bool listContains = stateCopy.data.any((element) {
              if (element == null) {
                return false;
              }
              return modelIdentifier.isIdentifier(
                model: element,
                identifier: param.id,
              );
            });
            if (listContains) {
              // Loaded objects contains the modified object.

              final crudGet = this.crudGet;
              if (crudGet == null) {
                // No read operation, can't update stateCopy.
                return stateCopy;
              }
              final newVersion = await crudGet.call(param.id);
              if (newVersion.isOk) {
                final newList = stateCopy.data.where(
                  (element) {
                    if (element == null) {
                      return false;
                    }
                    return !modelIdentifier.isIdentifier(
                      model: element,
                      identifier: param.id,
                    );
                  },
                ).toList();
                return CrudListLoaded<Model?>(newList + [newVersion.ok]);
              }
            }
            return stateCopy;
          }
          return const CrudUpdated();
        },
        (error) async => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> updateAll(Map<String, Object?> param) async {
    final crud = crudUpdateAll;
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudUpdating());
    final result = await crud.call(param);
    emit(
      await result.foldAsync(
        (_) async {
          if (stateCopy is CrudLoaded<Model?>) {
            // Same object, need to update actual stateCopy
            final crudGet = this.crudGet;
            if (crudGet == null) {
              // No read operation, can't update stateCopy.
              return stateCopy;
            }
            final actualId =
                modelIdentifier.getIdentifierOrNull(stateCopy.data);
            final newVersion = await crudGet.call(actualId ?? '');
            if (newVersion.isOk) {
              return CrudLoaded<Model?>(newVersion.ok);
            }
            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            final crudSearch = this.crudSearch;
            if (crudSearch == null) {
              // No read operation, can't update stateCopy.
              return stateCopy;
            }
            // Load all id to retrieve exactly same object
            // (not all because previous stateCopy can be a query result)
            final List<String?> ids = stateCopy.data
                .map(
                  (e) => modelIdentifier.getIdentifierOrNull(e),
                )
                .where((element) => element != null)
                .toList();
            final result = await crudSearch.call([
              WhereQuery(
                WhereQueryType.whereIn,
                'id',
                ids,
              ),
            ]);
            if (result.isOk) {
              return CrudListLoaded<Model?>(result.ok ?? []);
            }
            return stateCopy;
          }
          return const CrudUpdated();
        },
        (error) async => CrudError(error.toString()),
      ),
    );
  }
}
